package jump.libx.phase;

import jump.libx.phase.helper.ComparablePhase;
import jump.libx.phase.helper.MethodWrapper;
import jump.libx.phase.event.PhaseEventHandler;
import jump.libx.phase.event.PhaseListener;
import org.apache.log4j.Logger;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;

/**
 * Ein Phaser ist eine Klasse, die ein Zielobjekt vom Typ Phased zugewiesen bekommt, welches mit Phase Annotationen
 * ausgestattet ist. Daher muss das Zielobjekt mindestens das Marker Interface Phased implementieren,
 * sinnvoller ist jedoch ein daraus abgeleitetes Interface mit Methoden, die mit der Phase
 * Annotation versehen sind.
 * <br>
 * Der Phaser ist dann in der Lage, diese Methoden per Reflection auszulesen und in der richtigen
 * Reihenfolge aufzurufen.
 * <br>
 * Das Aufrufen erfolgt durch die verschiedenen invokeXxx() Methoden.
 * <pre>
 * Created by neh
 * Date: 30.03.2005
 * Time: 17:23:36
 * </pre>
 */
public class Phaser {

   private static Logger logger = Logger.getLogger( Phaser.class );
   private Phased _target;
   private PhaseEventHandler _eventHandler = null;
   private boolean _ordered = false;

   /**
    * Erwartet das Zielobjekt, welches aus dem Phased Interface eine Reihe von Methoden
    * enthalten sollte, die mit der Annotation Phase versehen sind.
    * @param target
    */
   public Phaser( Phased target ) {
      _target = target;
      _eventHandler = new PhaseEventHandler( this );
   }

   /**
    * Liefert das Zielobjekt dieses Phasers
    * @return
    */
   public Phased getPhased() {
      return _target;
   }

   /**
    * Liefert den PhaseEventHandler, der das Event Management dieses Phasers verwaltet.
    * @return
    */
   public PhaseEventHandler getEventHandler() {
      return _eventHandler;
   }

   /**
    * Delegiert den Aufruf an den PhaseEventHandler dieses Phasers weiter.
    * @param listener
    */
   public void addPhaseListener( PhaseListener listener ) {
      _eventHandler.addPhaseListener( listener );
   }

   /**
    * Delegiert den Aufruf an den PhaseEventHandler dieses Phasers weiter.
    * @param listener
    */
   public void removePhaseListener( PhaseListener listener ) {
      _eventHandler.removePhaseListener( listener );
   }

   /**
    * Liefert true, wenn der Phaser die Aufrufreihenfolge automatisch sortiert, false sonst.
    * @return
    */
   public boolean isOrdered() {
      return _ordered;
   }

   /**
    * Setzt das Verhalten des Phasers so, dass bei einem invoke Aufruf die bergebenenen Indizies bzw. Methodennamen
    * automatisch geordnet werden, falls ordered true ist. Andernfalls werden die Methoden in der bergebenen
    * Reihenfolge aufgerufen. Per default ist die Property ordered false.
    * @param ordered
    */
   public void setOrdered( boolean ordered ) {
      _ordered = ordered;
   }

   /**
    * Ruft alle Methoden aus dem Phased Interface auf, welche mit Phase Annotationen versehen sind.
    * @throws PhaserException
    */
   public void invokeAll() throws PhaserException {
      HashMap<Phase, Method> phaseMap = this.findPhasedMethods();
      ArrayList<MethodWrapper> methodList = this.orderMethodsByPhaseIndex( phaseMap );
      for ( MethodWrapper mw : methodList ) {
         _eventHandler.notifyBefore( mw.getPhase() );
         this.invokeMethod( mw.getMethod() );
         _eventHandler.notifyAfter( mw.getPhase() );
      }
   }

   /**
    * Ruft alle Methoden aus dem Phased Interface auf, welche mit Phase Annotationen versehen sind,
    * deren Phase Index in der Liste phaseList enthalten sind.
    * @param phaseList
    * @throws PhaserException
    */
   public void invoke( int[] phaseList ) throws PhaserException {
      HashMap<Phase, Method> phaseMap = this.findPhasedMethods();
      ArrayList<MethodWrapper> methodList = this.orderMethodsByPhaseIndex( phaseMap );
      for ( MethodWrapper mw : methodList ) {
         if ( phaseList == null || mw.isContained( phaseList ) ) {
            _eventHandler.notifyBefore( mw.getPhase() );
            this.invokeMethod( mw.getMethod() );
            _eventHandler.notifyAfter( mw.getPhase() );
         }
      }
   }

   /**
    * Ruft alle Methoden aus dem Phased Interface auf, welche mit Phase Annotationen versehen sind,
    * deren Methodennamen in der Liste methodNameList enthalten sind.
    * @param methodNameList
    * @throws PhaserException
    */
   public void invoke( String[] methodNameList ) throws PhaserException {
      HashMap<Phase, Method> phaseMap = this.findPhasedMethods();
      ArrayList<MethodWrapper> methodList = this.orderMethodsByPhaseIndex( phaseMap );
      for ( MethodWrapper mw : methodList ) {
         if ( methodNameList == null || mw.isContained( methodNameList ) ) {
            _eventHandler.notifyBefore( mw.getPhase() );
            this.invokeMethod( mw.getMethod() );
            _eventHandler.notifyAfter( mw.getPhase() );
         }
      }
   }

   /**
    * Ruft die bergebene Methode auf und kapselt ggf. entstehende Exceptions in PhaserExceptions.
    * @param method
    * @throws PhaserException
    */
   private void invokeMethod( Method method ) throws PhaserException {
      logger.debug( "Invoking: " + method.getName() );
      try {
         method.invoke( _target, new Object[]{} );
      } catch ( IllegalAccessException e ) {
         throw new PhaserException( e );
      } catch ( InvocationTargetException e ) {
         throw new PhaserException( e );
      }
   }

   /**
    * Findet die aufzurufenden Methoden in der Reihenfolge gemss der angebenenen Phase Index an den
    * Methoden
    * @param phaseMap
    * @return
    */
   private ArrayList<MethodWrapper> orderMethodsByPhaseIndex( HashMap<Phase, Method> phaseMap ) {
      ArrayList<ComparablePhase> phaseList = new ArrayList<ComparablePhase>();
      for ( Phase phase : phaseMap.keySet() ) {
         phaseList.add( new ComparablePhase( phase ) );
      }
      if ( _ordered ) {
         Collections.sort( phaseList );
      }
      ArrayList<MethodWrapper> orderedMethodList = new ArrayList<MethodWrapper>();
      for ( ComparablePhase cp : phaseList ) {
         Method method = phaseMap.get( cp.getPhase() );
         orderedMethodList.add( new MethodWrapper( method, cp.getPhase() ) );
      }
      return orderedMethodList;
   }

   /**
    * Liefert eine Abbildung von Phase auf Methode.
    * @return
    */
   private HashMap<Phase, Method> findPhasedMethods() {
      HashMap<Phase, Method> phaseMap = new HashMap<Phase, Method>();
      for ( Class clazz : _target.getClass().getInterfaces() ) {
         for ( Method method : clazz.getMethods() ) {
            if ( method.isAnnotationPresent( Phase.class ) ) {
               Phase phase = method.getAnnotation( Phase.class );
               phaseMap.put( phase, method );
            }
         }
      }
      for ( Method method : _target.getClass().getMethods() ) {
         if ( method.isAnnotationPresent( Phase.class ) ) {
            Phase phase = method.getAnnotation( Phase.class );
            phaseMap.put( phase, method );
         }
      }
      return phaseMap;
   }

}
